/*************************************************************************
 *  TinyFugue - programmable mud client
 *  Copyright (C) 1996, 1997, 1998, 1999, 2002, 2003, 2004, 2005, 2006-2007 Ken Keys
 *
 *  TinyFugue (aka "tf") is protected under the terms of the GNU
 *  General Public License.  See the file "COPYING" for details.
 ************************************************************************/
/* $Id: keylist.h,v 35000.16 2007/01/13 23:12:39 kkeys Exp $ */

gencode(CLEAR),
gencode(FLUSH),
gencode(LNEXT),
gencode(NEWLINE),
gencode(PAUSE),
gencode(RECALLB),
gencode(RECALLBEG),
gencode(RECALLEND),
gencode(RECALLF),
gencode(REDRAW),
gencode(REFRESH),
gencode(SEARCHB),
gencode(SEARCHF),
gencode(SELFLUSH)
